#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "gate.h"

int swap_system(char * cmd)
{ debug(7, "system: '%s'", cmd);
  return system(cmd);
}

int pipe_system0(int * in,int * out,char * cmd,char * argv0)
{ int hpipe[2];
  int newstdin=-1, newstdout=-1, pid;

  debug(7, "pipe_system: '%s'", cmd);
  fflush(stdout);
  fflush(stderr);
  if (in)
  { if (pipe(hpipe))
    { logwrite('?', "Can't create pipe: %s!\n", strerror(errno));
      return -1;
    }
    *in=hpipe[1];
    newstdin=hpipe[0];
  }
  if (out)
  { if (pipe(hpipe))
    { pid=errno;
      if (in)
      { close(*in);
        close(newstdin);
      }
      logwrite('?', "Can't create pipe: %s!\n", strerror(pid));
      return -1;
    }
    *out=hpipe[0];
    newstdout=hpipe[1];
  }
forkagain:
  pid=fork();
  if (pid==-1)
  { if (errno==EINTR || errno==EAGAIN)
      goto forkagain;
    pid=errno;
    if (in)
    { close(*in);
      close(newstdin);
    }
    if (out)
    { close(newstdout);
      close(*out);
    }
    logwrite('?', "Can't fork(): %s!\n", strerror(pid));
    return -1;
  }
  if (pid==0)
  { if (in)
    { close(*in);
      dup2(newstdin, fileno(stdin));
      close(newstdin);
    }
    if (out)
    { close(*out);
      dup2(newstdout, fileno(stdout));
      close(newstdout);
    }
    if (strpbrk(cmd, "<>|\"\'\\$*!?[]~`;()&")==0)
    { char *p=strpbrk(cmd, " \t");
      if (p) *p++='\0';
      else p=cmd+strlen(cmd);
      execlp(cmd, argv0 ? argv0 : cmd, p ? p : NULL, NULL);
      logwrite('?', "Can't exec %s: %s!\n", cmd, strerror(errno));
    }
    else
    { char *p=getenv("SHELL");
      if (p) p=strchr(p, '=');
      if (p) p++;
      else p="/bin/sh";
      execlp(p, p, "-c", cmd, NULL);
      logwrite('?', "Can't exec %s: %s!\n", strerror(errno));
    }
    exit(123);
  }
  if (out) close(newstdout);
  if (in) close(newstdin);
  debug(5, "pipe_system: \"%s\" pid %d", cmd, pid);
  return pid;
}

int pipe_spawnv(int * in,int * out,char * name, char * args[])
{
  int newstdin=-1, newstdout=-1;
  int pid, i;
  int hpipe[2];

  debug(7, "Pipe_Spawnv: %s", name);
  for (i=0;args[i];i++)
    debug(7, "Pipe_Spawnv: arg[%d]='%s'", i, args[i]);
  fflush(stdout);
  fflush(stderr);
  if (in)
  { if (pipe(hpipe))
    { logwrite('?', "Can't create pipe: %s!\n", strerror(errno));
      return -1;
    }
    *in=hpipe[1];
    newstdin=hpipe[0];
  }
  if (out)
  { if (pipe(hpipe))
    { pid=errno;
      if (in)
      { close(*in);
        close(newstdin);
      }
      logwrite('?', "Can't create pipe: %s!\n", strerror(pid));
      return -1;
    }
    *out=hpipe[0];
    newstdout=hpipe[1];
  }
forkagain1:
  pid=fork();
  if (pid==-1)
  { if (errno==EINTR || errno==EAGAIN)
      goto forkagain1;
    pid=errno;
    if (in)
    { close(*in);
      close(newstdin);
    }
    if (out)
    { close(newstdout);
      close(*out);
    }
    logwrite('?', "Can't fork(): %s!\n", strerror(pid));
    return -1;
  }
  if (pid==0)
  { if (in)
    { close(*in);
      dup2(newstdin, fileno(stdin));
      close(newstdin);
    }
    if (out)
    { close(*out);
      dup2(newstdout, fileno(stdout));
      close(newstdout);
    }
    execvp(name, args);
    logwrite('?', "Can't exec %s: %s!\n", name, strerror(errno));
    exit(123);
  }
  if (out) close(newstdout);
  if (in) close(newstdin);
  debug(5, "Pipe_Spawnv: %s pid %d", name, pid);
  return pid;
}
