#include <stdio.h>
#include "gate.h"

#define EX_OK           0       /* successful termination */

#define EX__BASE        64      /* base value for error messages */

#define EX_USAGE        64      /* command line usage error */
#define EX_DATAERR      65      /* data format error */
#define EX_NOINPUT      66      /* cannot open input */
#define EX_NOUSER       67      /* addressee unknown */
#define EX_NOHOST       68      /* host name unknown */
#define EX_UNAVAILABLE  69      /* service unavailable */
#define EX_SOFTWARE     70      /* internal software error */
#define EX_OSERR        71      /* system error (e.g., can't fork) */
#define EX_OSFILE       72      /* critical OS file missing */
#define EX_CANTCREAT    73      /* can't create (user) output file */
#define EX_IOERR        74      /* input/output error */
#define EX_TEMPFAIL     75      /* temp failure; user is invited to retry */
#define EX_PROTOCOL     76      /* remote error in protocol */
#define EX_NOPERM       77      /* permission denied */
#define EX_CONFIG       78      /* configuration error */

#define EX__MAX         78

static char *sendmailerr[] = {
"command line usage error",
"data format error",
"cannot open input",
"addressee unknown",
"host name unknown",
"service unavailable",
"internal software error",
"system error",
"critical OS file missing",
"can't create (user) output file",
"input/output error",
"temp failure; user is invited to retry",
"remote error in protocol",
"permission denied",
"configuration error"
};

char *sys_siglist[] = {
"0",
"SIGHUP",
"SIGINT",
"SIGQUIT",
"SIGILL",
"SIGTRAP",
"SIGABRT",
"SIGIOT",
"SIGBUS",
"SIGFPE",
"SIGKILL",
"SIGUSR1",
"SIGSEGV",
"SIGUSR2",
"SIGPIPE",
"SIGALRM",
"SIGTERM",
"SIGSTKFLT",
"SIGCLD",
"SIGCONT",
"SIGSTOP",
"SIGTSTP",
"SIGTTIN",
"SIGTTOU",
"SIGURG",
"SIGXCPU",
"SIGXFSZ",
"SIGVTALRM",
"SIGPROF",
"SIGWINCH",
"SIGIO",
"SIGPWR",
"SIGUNUSED"
};

char * strsysexit(int retcode)
{
  static char unknownerr[128];

  if (retcode>=EX__BASE &&
      retcode<=EX__MAX)
    return sendmailerr[retcode-EX__BASE];
  sprintf(unknownerr, "retcode %u", retcode);
  return unknownerr;
}

char * strsignal(int signo)
{
  static char unknownsig[20];

  if (signo>=0 && signo<sizeof(sys_siglist)/sizeof(sys_siglist[0]))
    return sys_siglist[signo];
  sprintf(unknownsig, "%u", signo);
  return unknownsig;
}
